/*
 *  CVI
 *
 *  cvi (change-vi) adds change-bars when updating documents.  This program
 *  also keeps track of updates to those files.  Initially, a user may use
 *  "vi" or "cvi" to edit a file.  cvi can then be used on subsequent edits 
 *  if the original file needs to be updated.  After editing, the updated 
 *  file is saved with the original file name -and contains change-bars.  
 *  The original file is saved with a revision number appended to the file 
 *  name.  Thus, a user can keep a history of all changes made to a file.
 *
*/

#include <stdio.h>
#include <sys/stat.h>

#define  SLEN	256			/* buffer size                        */
#define  IFDIR  0040000                 /* for stat command                   */

char	 bf1[SLEN],			/* work space                         */
    	 bf2[SLEN];			/* work space                         */


/*
 *  Main.
 *
 *  Get command line argument (the file name).  If the file does not exist, 
 *  call edit() to put the user into vi.  If the file exists, call change()
 *  to update the new file with change-bars and save a copy of the old file.
 *
*/
main(argc, argv)
int argc;
char **argv;
{

  char new[SLEN];		 /* original file to be changed -from argv[1] */
  int  pid;


  pid = getpid();

  switch (argc) {
  case 2:
    if (argv[1][0] == '-') {
      printf("Usage: cvi [-v] file_name\n");
      exit(0);
    }
    strcpy(new, argv[1]);
    break;
  case 3:
    if (!equal(argv[1], "-v")) {
      printf("Usage: cvi [-v] file_name\n");
      exit(0);
    }
    strcpy(new, argv[2]);
    break;
  default:
    printf("Usage: cvi [-v] file_name\n");
    exit(0);
  }

  switch (filesize(new)) {
  case -1:
    edit(new);
    break;
  case -2:
    printf("Cannot edit a directory!!\n");
    exit(0);
    break;
  default:
      if (equal(argv[1], "-v"))
        change(new, pid, 1);
      else
        change(new, pid, 0);
  }
}


/*
 *  Change.
 *
 *  Change is called to update an existing file with new information.
 *  The updated file will retain the original file name while the original
 *  file is saved with a revision number appended to the file name.
 *
*/
change(new, pid, vopt)
char *new;
int pid, vopt;
{
  FILE *chd, *cmd, *nfd, *ofd;
  char chg[SLEN],				/* CHG file's name buffer     */
       chm[SLEN],				/* CHM file's name buffer     */
       org[SLEN];				/* ORG file's name buffer     */
  int dot, i, last, multi;


  /*
   *  Assign the names of our temporary files now.  "chg" will hold diff(1m)
   *  changes.  "chm" will hold modified diff(1m) changes.
   *
  */
  sprintf(chg, "/tmp/chg.%d", pid);
  sprintf(chm, "/tmp/chm.%d", pid);

  /*  
   *  We have to compare the new file to the original file.  Therefore, we
   *  copy "new" file to the same file name, plus a revision number.  This
   *  becomes the "org"inal file.  The user will work on the "new" file.
   *
  */
  sprintf(org, "%s.%d", new, getrev(new));
  copy(new, org);

  /*
   *  Allow the user to change the new file -using the vi editor.  The new
   *  file retains the original file name, but the original file gets a
   *  revision number.
   *
  */
  edit(new);
  printf("Original file renamed \"%s\".\n", org);
  printf("Adding change-bars to file \"%s\"...", new);
  fflush(stdout);

  /*
   *  Run diff(1m), using the -e option, to create an editing script.  This
   *  script will hold the required changes that could convert the original
   *  file to the new file.
   *
  */
  differentiate(org, new, chg);
  
  /*
   *  Fix up the editing script created by differentiate() so that change-
   *  bars will be added.  Open "chg" for reading -we will read the necessary 
   *  edit requirements from this file.  Open "chm" file for writing -we will 
   *  put the modified editing requirements for ed(1m) in this file.  At the
   *  "while" loop, we read the editing requirements from chg, modify them,
   *  and write them into the chm file.
   *
  */
  if ((chd = fopen(chg, "r")) == NULL) {
    printf("fopen failed for %s", chg);
    cleanup(chg, chm, pid);
  }

  if ((cmd = fopen(chm, "w")) == NULL) {
    printf("fopen failed for %s", chm);
    cleanup(chg, chm, pid);
  }

  dot = 1;
  while (fgets(bf1, SLEN, chd) != NULL) {
    multi = 0;
    if (bf1[0] == '.' && strlen(bf1) == 2) {
      fprintf(cmd, "%s", bf1);
      dot = 1;
    }
    else if (!numeric(bf1[0])) {
      if (strlen(bf1) > 1) {
        strcpy(bf2, "|");
        strcat(bf2, bf1);
        fprintf(cmd, "%s", bf2);
	bf2[0] = '\0';
      }
      else
        fprintf(cmd, "%s", bf1);
    }
    else {  /* numeric */
      last = strlen(bf1) - 2;
      switch (bf1[last]) {                      
      case 'a': 
      case 'c': 
	if (dot == 1) {
	  dot = 0;
          fprintf(cmd, "%s", bf1);
          if (bf1[last] == 'c' && vopt) {
            for (i = 0; i < last; i++)
	      if (bf1[i] == ',') {
	        bf1[last] = '\0';
	        sprintf(bf2, 
		  "| WARNING!  Lines May Have Been Deleted (From,To: %s)\n",
		  bf1);
	        fprintf(cmd, "%s", bf2);
	        break;
              }
          }
        }
	else {
          if (strlen(bf1) > 1) {
            strcpy(bf2, "|");
            strcat(bf2, bf1);
            fprintf(cmd, "%s", bf2);
	    bf2[0] = '\0';
          }
          else
            fprintf(cmd, "%s", bf1);
        }
	break;
      case 'd':
        fprintf(cmd, "%s", bf1);
	if (vopt) {
          for (i = 0; i < last; i++)
	    if (bf1[i] == ',') {
	      multi = 1;
	      break;
            }
	  bf1[last] = '\0';
          if (multi)
	    sprintf(bf2, "i\n| DELETED LINES (From,To: %s)\n.\n", bf1);
	  else 
	    sprintf(bf2, "i\n| DELETED LINE Number: %s\n.\n", bf1);
	  fprintf(cmd, "%s", bf2);
	}
	break;
      default:
        if (strlen(bf1) > 1) {
          strcpy(bf2, "|");
          strcat(bf2, bf1);
          fprintf(cmd, "%s", bf2);
	  bf2[0] = '\0';
        }
        else
          fprintf(cmd, "%s", bf1);
      }
    }
  }
   
  fflush(cmd);
  fclose(cmd);
  fclose(chd);

  /*
   *  Now, implement the changes to the original file.  The "chm" file
   *  contains the edit requirements with the change-bar modifications.  The
   *  "org" file is the original file -we save that file for the user with
   *  a revision number.  The "new" file has the previous name of the
   *  "org" file.
   *
  */
  ed(chm, org, new);

  printf("done.\n");

  cleanup(chg, chm, pid);
}


/* 
 *  Cleanup.
 *
 *  Force-remove the change file, then the modified change file, and then
 *  exit gracefully!
 *
*/
cleanup(chg, chm, pid)
{
  sprintf(bf1, "rm -f %s", chg);
  uxecute(bf1);

  sprintf(bf1, "rm -f %s", chm);
  uxecute(bf1);

  exit(0);
}


/*
 *  Copy.
 *
 *  Copy "new" to "org".
 *
*/
copy(new, org)
char *new, *org;
{
  sprintf(bf1, "cp %s %s", new, org);
  uxecute(bf1);
}


/*
 *  Differentiate.
 *
 *  Differentiate file org and new with the -e option to create an ed(1m)
 *  script.  Store the script of required ed(1m) changes in chg.
 *
*/
differentiate(org, new, chg)
char *org, *new;
{
  sprintf(bf1, "diff -e %s %s > %s", org, new, chg);
  uxecute(bf1);
}


/*
 *  Ed.
 *
 *  Run ed(1m) script to update original file with changes from the ed(1m)
 *  script created by diff(1m).  Put the changed file (with change-bars)
 *  in the original file name -overwriting the original file.
 *
*/
ed(chm, org, new)
char *chm, *org, *new;
{
  sprintf(bf1, "(cat %s;echo '1,$p') | ed - %s > %s", chm, org, new);
  uxecute(bf1);
}


/*
 *  Edit.
 *
 *  Edit simply arranges a buffer to contain the "vi" command along with
 *  the input file that was passed in.  Then edit calls uxecute so that
 *  the "new" file may be edited.
 *
*/
edit(new)
char *new;
{
  sprintf(bf1, "vi %s", new);
  uxecute(bf1);
}


/* 
 *  Filesize.
 *
 *  Returns -1 if no file, -2 if directory, or the number of bytes in
 *  the specified filename.
 *
*/
int filesize(filename)
char *filename;
{
  struct stat statbf1;

  if (stat(filename, &statbf1) != 0)
    return(-1);
  else if (statbf1.st_mode & IFDIR)
    return(-2);
  else
    return((int) statbf1.st_size);
}


/* 
 *  Getrev.
 *
 *  This function looks to see if the original file has a .N number.  If so,
 *  we figure out what the next sequence is so that we can name our "org"
 *  file something like "org.N+1".  This allows the user to keep multiple
 *  changes to an original file.
 *
*/
getrev(new)
char *new;
{
  char num[SLEN];				/* Revision number of file    */
  int rev = 0;

  do {
    rev++;
    sprintf(num, "%s.%d", new, rev);
  } while (filesize(num) > 0);

  return(rev);
}


/* 
 *  Numeric.
 *
 *  This function returns 1 if the character passed in is a number, 0
 *  if not.
 *
*/
numeric(c)
char c;
{
  if (c >= '0' && c <= '9')
    return(1);
  else
    return(0);
}


/* 
 *  Uxecute.
 *
 *  This function forks a shell and runs UNIX commands for cvi.
 *
*/
uxecute(bf1)
char *bf1;
{
  int pid, status;

  switch (pid = vfork()) {
  case -1:
    printf("sh: fork failed...\n");
    break;
  case  0:
    execl("/bin/sh", "sh", "-c", bf1, (char *)0);
    break;
  default:
    wait(&status);
  }
}


/* 
 *  Equal.
 *
 *  Returns 1 if strings a and b are equal. 
 *
*/
equal(a,b)
char *a, *b;
{		
  if (strcmp(a,b) == 0)
    return(1);
  else
    return(0);
}

